import os
import sys

import psutil

from monk.keras_prototype import prototype
from monk.compare_prototype import compare
from monk.pip_functionality_tests.keras.common import print_start
from monk.pip_functionality_tests.keras.common import print_status



def test_expert_train(system_dict):
    forward = True;
    if(not os.path.isdir("datasets")):
        os.system("! wget --load-cookies /tmp/cookies.txt \"https://docs.google.com/uc?export=download&confirm=$(wget --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2' -O- | sed -rn 's/.*confirm=([0-9A-Za-z_]+).*/\1\n/p')&id=1rG-U1mS8hDU7_wM56a1kc-li_zHLtbq2\" -O datasets.zip && rm -rf /tmp/cookies.txt")
        os.system("! unzip -qq datasets.zip")

    test = "expert_train_object_creation";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf = prototype(verbose=0);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Prototype()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Prototype("sample-project-1", "sample-experiment-4");
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Dataset_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Dataset_Params(dataset_path=["datasets/dataset_csv_id/train", "datasets/dataset_csv_id/val"],
                path_to_csv=["datasets/dataset_csv_id/train.csv", "datasets/dataset_csv_id/val.csv"],
                split=0.9, input_size=224, batch_size=2, shuffle_data=True, num_processors=3);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Dataset()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Dataset();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");


    test = "expert_train_Model_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Model_Params(model_name="resnet50", freeze_base_network=True, use_gpu=True, gpu_memory_fraction=0.5, use_pretrained=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");

    
    test = "expert_train_append_layer()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.append_dropout(probability=0.1);
            ktf.append_linear(final_layer=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");            


    test = "expert_train_Model()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Model();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 




    test = "expert_train_lr_step_decrease()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.lr_step_decrease(1, gamma=0.9);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 


    test = "expert_train_optimizer_sgd()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.optimizer_sgd(0.0001, momentum=0.9);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 



    test = "expert_train_loss_softmax_crossentropy()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.loss_crossentropy();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped"); 

    
    test = "expert_train_Training_Params()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Training_Params(num_epochs=3, display_progress=True, display_progress_realtime=True, 
                save_intermediate_models=True, intermediate_model_prefix="intermediate_model_", save_training_logs=True);
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");




    test = "expert_train_Train()";
    system_dict["total_tests"] += 1;
    print_start(test, system_dict["total_tests"])
    if(forward):
        try:
            ktf.Train();
            system_dict["successful_tests"] += 1;
            print_status("Pass");
        except Exception as e:
            system_dict["failed_tests_exceptions"].append(e);
            system_dict["failed_tests_lists"].append(test);
            forward = False;
            print_status("Fail");
    else:
        system_dict["skipped_tests_lists"].append(test);
        print_status("Skipped");        
    


    return system_dict



from monk.tf_keras_1.models.imports import *


class CNNVisualizer():
  '''
  Base class for visualizing features and feature maps of keras models
  
  Args:
        model (): Trained model 
        is_notebook: Check environment type
                      True - Using Jupyter-like environment (Notebook)
                      False - Otherwise 
  '''

  def __init__(self, model, is_notebook=False):
    self.model = model
    self.conv_layers = self._extract_conv_layers()
    self.fmaps = []
    self.is_notebook = is_notebook

  def _extract_conv_layers(self):

    '''
    Extracts convolutional layers from a model 
    '''

    conv_layers = {}
    num_layers = 0
    for layer in self.model.layers:
        if type(layer) == keras.layers.convolutional.Conv2D:
          num_layers += 1
          conv_layers.update({layer.name : layer})
      
    print("Found {} convolutional layers".format(num_layers))
    return conv_layers

  def get_layer_info(self, layer_name=None):

    '''
    Displays imformation of a layer 

    Args:
        layer_name (str): Name of the layer whose information is needed
    '''

    name = layer_name
    layer = self.conv_layers[layer_name]

    num_filters = layer.kernel.shape[3]
    filter_w = layer.kernel.shape[0]
    filter_h = layer.kernel.shape[1]
    num_channels = layer.kernel.shape[2]
    
    print("Found: Layer name: {} | Kernel info: {} filters of size ({} X {}) with {} channels".format(name,num_filters,filter_w,filter_h,num_channels))
    
  def get_feature_maps(self, image_path):

    '''
    Obtain feature maps generated on an image 

    Args:
        image_path (str): Path of the image whose feature maps are needed
    '''

    target_size = (self.model.layers[0].input.shape[1], self.model.layers[0].input.shape[2])
    img = load_img(image_path, target_size=target_size)
    img = img_to_array(img)
    img = np.expand_dims(img, axis=0)

    outputs = [layer.output for layer in self.conv_layers.values()]
    model_temp = Model(inputs=self.model.inputs, outputs=outputs)
    self.fmaps = model_temp.predict(img)

  def visualize_kernel(self, layer_name='conv_1d',filter_id=0, channel_id=0, num_channels=1, cmap='gray', return_kernel=False):

    '''
    Visualize kernel weights of a given layer 

    Args:
        layer_name (str): Name of the layer whose feature map is needed
        filter_id (int): Filter that needs to be visualized
        channel_id (int): Channel that needs to be visualized
        num_channels (int): Number of channels to be displayed at once
        cmap (str): CMAP used for displaying image
        return_kernel (bool): Whether the kernel needs to be returned
    
    Returns:
        Kernel specified by the paramters if return_kernel is True.
    '''

    layer = self.conv_layers[layer_name]
    self.get_layer_info(layer_name)

    kernels = layer.get_weights()[0]

    if filter_id >= layer.kernel.shape[3] or channel_id >= layer.kernel.shape[2]:
       return
    
    if num_channels == 1:
      kernel = kernels[:,:,:,filter_id][:,:,channel_id]
    else:
      kernel = kernels[:,:,:,filter_id][:,:,channel_id:(channel_id+num_channels)]

    if (kernel.max() - kernel.min()) != 0:
      kernel = (kernel - kernel.min())/(kernel.max()-kernel.min())
    
    if return_kernel:
      return kernel
    
    num_rows = int(num_channels/4)
    num_cols = min(4, int(layer.kernel.shape[2]))

    print("")
    
    fig = plt.figure(figsize=(4*num_cols,4*num_rows))

    gs = gridspec.GridSpec(num_rows, num_cols)
    
    fig.subplots_adjust(top=0.88)
    fig.suptitle("Filter id: "+str(filter_id), fontsize=16)
    fig.tight_layout()

    for i in range(kernel.shape[2]):
      channel = kernel[:,:,i]
      ax = plt.subplot(gs[i])
      plt.axis('on')
      ax.set_title("Channel id: "+str(channel_id+i),wrap=True)
      ax.set_xticklabels([])
      ax.set_yticklabels([])
      ax.set_aspect('equal')
      ax.imshow(channel, cmap=cmap)

    plt.show()

    if num_channels == 1:
      plt.title("Filter id: "+str(filter_id)+" | Channel id: "+str(channel_id),wrap=True)
      plt.imshow(kernel,cmap=cmap)
      plt.show()

  def visualize_image(self, layer_name='conv_1d', channel_id=0, num_channels=1, cmap='gray', return_fmap=False):

    '''
    Visualize a feature map given an image 

    Args:
        layer_name (str): Name of the layer whose feature map is needed
        channel_id (int): Channel that needs to be visualized
        num_channels (int): Number of channels to be displayed at once
        cmap (str): CMAP used for displaying image
        return_fmap (bool): Whether the feature map needs to be returned
    
    Returns:
        Feature map specified by the paramters if return_fmap is True.
    '''

    layer = self.conv_layers[layer_name]
    self.get_layer_info(layer_name)
    
    if channel_id >= layer.output.shape[3]:
        return
    
    layer_id = list(self.conv_layers.keys()).index(layer_name)

    if num_channels == 1:
      fmap = self.fmaps[layer_id][0,:,:,channel_id]
    else:
      fmap = self.fmaps[layer_id][0,:,:,channel_id:(channel_id+num_channels)]
    
    if return_fmap:
      return fmap

    num_rows = int(num_channels/4)
    num_cols = min(4, layer.output.shape[3])

    print("")
    
    fig = plt.figure(figsize=(4*num_cols,4*num_rows))

    gs = gridspec.GridSpec(num_rows, num_cols)
    
    fig.subplots_adjust(top=0.88)
    fig.tight_layout()

    for i in range(fmap.shape[2]):
      channel = fmap[:,:,i]
      ax = plt.subplot(gs[i])
      plt.axis('on')
      ax.set_title("Filter id: "+str(channel_id+i),wrap=True)
      ax.set_xticklabels([])
      ax.set_yticklabels([])
      ax.set_aspect('equal')
      ax.imshow(channel, cmap=cmap)

    plt.show()

    if num_channels == 1:
      plt.title("Filter id: "+str(channel_id),wrap=True)
      plt.imshow(fmap,cmap=cmap)
      plt.show() 

  def store_kernels(self, path='kernels', cmap='gray'):

    '''
    Store kernel weights as images 

    Args:
        path (str): Path to the directory where the images need to be stored
        cmap (str): The CMAP used for the images
    '''

    for layer_name, layer in self.conv_layers.items():
      layer_path = os.path.join(path, layer_name) 
      os.mkdir(layer_path)
      
      num_filters = layer.kernel.shape[3]
      num_channels = layer.kernel.shape[2]
      
      for f_id in range(num_filters):
        f_path = os.path.join(layer_path, 'feature_'+str(f_id))
        os.mkdir(f_path)
        for ch_id in range(num_channels):
          kernel = self.visualize_kernel(layer_name=layer_name,filter_id=f_id, channel_id=ch_id, cmap=cmap, return_kernel=True)
          fig_path = f_path + '/channel_' + str(ch_id) + '.png'
          plt.imsave(fname=fig_path, arr=kernel, cmap=cmap)

  def store_feature_maps(self, path='fmaps', cmap='gray'):

    '''
    Stores the feature maps generated by the model on an input image 

    Args:
        path (str): Path to the directory where the feature maps need to be stored
        cmap (str): The CMAP used for the images
    '''

    for layer_name, layer in self.conv_layers.items():
      layer_path = os.path.join(path, layer_name) 
      os.mkdir(layer_path)
      
      num_channels = layer.output.shape[3]
      
      for ch_id in range(num_channels):
        fmap = self.visualize_image(layer_name=layer_name, channel_id=ch_id, cmap=cmap, return_fmap=True)
        fig_path = layer_path + '/channel_' + str(ch_id) + '.png'
        plt.imsave(fname=fig_path, arr=fmap, cmap=cmap)

  def handle_layer_change(self,change):
    clear_output()

    layer = self.conv_layers[change.new]
    self.filters.value = 0
    self.channels.value = 0
    self.filters.max = (layer.kernel.shape[3]-1)
    self.channels.max = (layer.kernel.shape[2]-1)
    self.num_channels.value = 4
    
    interact(self.visualize_kernel,layer_name=self.layers_by_name, filter_id=self.filters, channel_id=self.channels, num_channels=self.num_channels, cmap=self.cmap, return_kernel=fixed(False))
    
  def handle_num_channels_change(self, change):
    self.channels.step = change.new
    
  def handle_layer_change_image(self,change):
    clear_output()

    layer = self.conv_layers[change.new]
    self.filters.value = 0
    self.filters.max = (layer.output.shape[3]-1)
    self.num_channels.value = 4

    interact(self.visualize_image,layer_name=self.layers_by_name, channel_id=self.filters, num_channels=self.num_channels, cmap=self.cmap, return_fmap=fixed(False))
  
  def handle_num_channels_change_image(self, change):
    self.filters.step = change.new
    
  def visualize_kernels(self, store_path=None):

    '''
    Visualize weights of kernels in the model

    Args:
        store_path (str): Path to the directory where the feature visuals need to be stored

    Returns:
        Interactive Ipython widget displaying kernel weights.
    '''

    if (not self.is_notebook) or (store_path is not None): 
      self.store_kernels(path=store_path)
      return

    self.layers_by_name = widgets.Dropdown(options=self.conv_layers.keys(), value=list(self.conv_layers.keys())[0], description='Layer')
    layer = self.conv_layers[self.layers_by_name.value]
    self.filters = widgets.IntSlider(value=0, min=0, max=(layer.kernel.shape[3]-1), step=1, description='Filter')
    self.channels = widgets.IntSlider(value=0, min=0, max=(layer.kernel.shape[2]-1), step=4, description='Channel')
    self.num_channels = widgets.IntSlider(value=4, min=4, max=20, step=4, description='# Channels')
    self.cmap = widgets.Dropdown(options=['viridis','gray'], value='gray', description='CMAP')
    self.layers_by_name.observe(self.handle_layer_change, 'value')
    self.num_channels.observe(self.handle_num_channels_change, 'value')
    
    interact(self.visualize_kernel,layer_name=self.layers_by_name, filter_id=self.filters, channel_id=self.channels, num_channels=self.num_channels, cmap=self.cmap, return_kernel=fixed(False))

  def visualize_feature_maps(self, image_path, store_path=None):

    '''
    Visualize the feature maps given an image 

    Args:
        image_path (str): Path to the image
        store_path (str): Path to the directory where the feature maps need to be stored

    Returns:
        Interactive Ipython widget displaying feature maps.
    '''
    
    self.get_feature_maps(image_path)

    if (not self.is_notebook) or (store_path is not None): 
      self.store_feature_maps(path=store_path)
      return
    
    self.layers_by_name = widgets.Dropdown(options=self.conv_layers.keys(), value=list(self.conv_layers.keys())[0], description='Layer')
    layer = self.conv_layers[self.layers_by_name.value]
    self.num_channels = widgets.IntSlider(value=4, min=4, max=20, step=4, description='# Filters')
    self.filters = widgets.IntSlider(value=0, min=0, max=(layer.output.shape[3]-1), step=4, description='Filter')
    self.cmap = widgets.Dropdown(options=['viridis','gray'], value='gray', description='CMAP')
    
    self.layers_by_name.observe(self.handle_layer_change_image, 'value')
    self.num_channels.observe(self.handle_num_channels_change_image, 'value')
    
    interact(self.visualize_image,layer_name=self.layers_by_name, channel_id=self.filters, num_channels=self.num_channels, cmap=self.cmap, return_fmap=fixed(False))


from monk.pytorch.finetune.imports import *
from monk.system.imports import *
from monk.pytorch.finetune.level_2_model_base import finetune_model



class finetune_training(finetune_model):
    '''
    Base class for training and associated functions

    Args:
        verbose (int): Set verbosity levels
                        0 - Print Nothing
                        1 - Print desired details
    '''
    @accepts("self", verbose=int, post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def __init__(self, verbose=1):
        super().__init__(verbose=verbose);



    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def get_training_estimate(self):
        '''
        Get estimated time for training a single epoch based on all set parameters

        Args:
            None

        Returns:
            float: Total time per epoch in seconds
        '''
        total_time_per_epoch = 0;

        self.system_dict = load_optimizer(self.system_dict);
        self.system_dict = load_scheduler(self.system_dict);
        self.system_dict = load_loss(self.system_dict);

        since = time.time();

        

        for phase in ['train', 'val']:
            if phase == 'train':
                self.system_dict["local"]["model"].train() 
            else:
                self.system_dict["local"]["model"].eval()  

            running_loss = 0.0
            running_corrects = 0

            required_iters = len(self.system_dict["local"]["data_loaders"][phase])//10;
            current_iter = 0;

            for inputs, labels in self.system_dict["local"]["data_loaders"][phase]:

                inputs = inputs.to(self.system_dict["local"]["device"]);
                labels = labels.to(self.system_dict["local"]["device"]);

                self.system_dict["local"]["optimizer"].zero_grad();


                with torch.set_grad_enabled(phase == 'train'):
                    if(self.system_dict["model"]["params"]["model_name"]):
                        if "inception" in self.system_dict["model"]["params"]["model_name"] and phase == 'train':
                            outputs, aux_outputs = self.system_dict["local"]["model"](inputs)
                            loss1 = self.system_dict["local"]["criterion"](outputs, labels)
                            loss2 = self.system_dict["local"]["criterion"](aux_outputs, labels)
                            loss = loss1 + 0.4*loss2
                        else:
                            outputs = self.system_dict["local"]["model"](inputs)
                            loss = self.system_dict["local"]["criterion"](outputs, labels)
                    else:
                        outputs = self.system_dict["local"]["model"](inputs)
                        loss = self.system_dict["local"]["criterion"](outputs, labels)

                    _, preds = torch.max(outputs, 1)

                    if phase == 'train':
                        loss.backward()
                        self.system_dict["local"]["optimizer"].step()

                running_loss += loss.item() * inputs.size(0)
                running_corrects += torch.sum(preds == labels.data)

                current_iter += 1;
                if(current_iter >= required_iters):
                    break;




        total_time_per_epoch = (time.time() - since)*10;
        
        return total_time_per_epoch;

    ###############################################################################################################################################





    ###############################################################################################################################################
    @accepts("self", post_trace=False)
    #@TraceFunction(trace_args=True, trace_rv=True)
    def set_training_final(self):
        '''
        Main training function

        Args:
            None

        Returns:
            None
        '''
        if(self.system_dict["states"]["resume_train"]):
            self.custom_print("Training Resume");
            total_time_per_epoch = 0;

            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_loss(self.system_dict);
            
            self.system_dict["training"]["status"] = False;

            pid = os.getpid();

            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                val_acc_history = list(np.load(self.system_dict["log_dir"] + "val_acc_history.npy", allow_pickle=True));
                train_acc_history = list(np.load(self.system_dict["log_dir"] + "train_acc_history.npy", allow_pickle=True));
                val_loss_history = list(np.load(self.system_dict["log_dir"] + "val_loss_history.npy", allow_pickle=True));
                train_loss_history = list(np.load(self.system_dict["log_dir"] + "train_loss_history.npy", allow_pickle=True));

            best_acc = 0.0;
            best_acc_epoch = 0;
            max_gpu_usage = 0;
            best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());

            for epoch in range(self.system_dict["hyper-parameters"]["num_epochs"]):
                if(self.system_dict["training"]["settings"]["display_progress"]):
                    self.custom_print('    Epoch {}/{}'.format(epoch+1, self.system_dict["hyper-parameters"]["num_epochs"]))
                    self.custom_print('    ' + '-' * 10)


                if(epoch < self.system_dict["training"]["outputs"]["epochs_completed"]):
                    self.custom_print("Skipping Current Epoch");
                    self.custom_print("");
                    self.custom_print("");
                    continue;

                since = time.time();

                for phase in ['train', 'val']:
                    if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                        pbar=tqdm(total=len(self.system_dict["local"]["data_loaders"][phase]));

                    if phase == 'train':
                        self.system_dict["local"]["model"].train() 
                    else:
                        self.system_dict["local"]["model"].eval()  

                    running_loss = 0.0
                    running_corrects = 0
                    total_labels = 0


                    for inputs, labels in self.system_dict["local"]["data_loaders"][phase]:
                        if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                            pbar.update();
                        inputs = inputs.to(self.system_dict["local"]["device"]);
                        labels = labels.to(self.system_dict["local"]["device"]);

                        self.system_dict["local"]["optimizer"].zero_grad();


                        with torch.set_grad_enabled(phase == 'train'):
                            if(self.system_dict["model"]["params"]["model_name"]):
                                if "inception" in self.system_dict["model"]["params"]["model_name"] and phase == 'train':
                                    outputs, aux_outputs = self.system_dict["local"]["model"](inputs)
                                    loss1 = self.system_dict["local"]["criterion"](outputs, labels)
                                    loss2 = self.system_dict["local"]["criterion"](aux_outputs, labels)
                                    loss = loss1 + 0.4*loss2
                                else:
                                    outputs = self.system_dict["local"]["model"](inputs)
                                    loss = self.system_dict["local"]["criterion"](outputs, labels)
                            else:
                                outputs = self.system_dict["local"]["model"](inputs)
                                loss = self.system_dict["local"]["criterion"](outputs, labels)

                            _, preds = torch.max(outputs, 1)

                            if phase == 'train':
                                loss.backward()
                                self.system_dict["local"]["optimizer"].step()


                        running_loss += loss.item() * inputs.size(0)
                        if(self.system_dict["dataset"]["label_type"] == "single"):
                            running_corrects += torch.sum(preds == labels.data)
                        else:
                            labels = labels.cpu().detach().numpy()[0]
                            list_classes = [];
                            list_labels = [];
                            raw_scores = outputs.cpu().detach().numpy()[0];

                            for i in range(len(raw_scores)):
                                prob = logistic.cdf(raw_scores[i])
                                if(prob > 0.5):
                                    list_classes.append(self.system_dict["dataset"]["params"]["classes"][i])

                            for i in range(len(labels)):
                                if(labels[i]):
                                    list_labels.append(self.system_dict["dataset"]["params"]["classes"][i])

                            for i in range(len(list_labels)):
                                actual = list_labels[i];
                                if actual in list_classes:
                                    correct = True;
                                else:
                                    correct = False;

                                index = self.system_dict["dataset"]["params"]["classes"].index(actual);
                                total_labels += 1;

                                if(correct):
                                    running_corrects += 1;


                    epoch_loss = running_loss / len(self.system_dict["local"]["data_loaders"][phase].dataset)

                    if(self.system_dict["dataset"]["label_type"] == "single"):
                        epoch_acc = running_corrects.double() / len(self.system_dict["local"]["data_loaders"][phase].dataset)
                    else:
                        epoch_acc = running_corrects/total_labels;


                    if(not os.getcwd() == "/kaggle/working"):
                        if(self.system_dict["model"]["params"]["use_gpu"]):
                            GPUs = GPUtil.getGPUs()
                            gpuMemoryUsed = GPUs[0].memoryUsed
                            if(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] < int(gpuMemoryUsed)):
                                self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = int(gpuMemoryUsed);
                    else:
                        gpuMemoryUsed = 0;
                        self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;



                    if(self.system_dict["training"]["settings"]["save_training_logs"]):
                        if(self.system_dict["dataset"]["label_type"] == "single"):
                            if phase == 'val':
                                val_acc = epoch_acc;
                                val_loss = epoch_loss;
                                val_acc_history.append(epoch_acc.cpu().detach().numpy());
                                val_loss_history.append(epoch_loss);
                            else:
                                train_acc = epoch_acc;
                                train_loss = epoch_loss;
                                train_acc_history.append(epoch_acc.cpu().detach().numpy());
                                train_loss_history.append(epoch_loss);
                        else:
                            if phase == 'val':
                                val_acc = epoch_acc;
                                val_loss = epoch_loss;
                                val_acc_history.append(epoch_acc);
                                val_loss_history.append(epoch_loss);
                            else:
                                train_acc = epoch_acc;
                                train_loss = epoch_loss;
                                train_acc_history.append(epoch_acc);
                                train_loss_history.append(epoch_loss);

                if(self.system_dict["training"]["settings"]["save_intermediate_models"]):
                    torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + 
                        self.system_dict["training"]["settings"]["intermediate_model_prefix"] + "{}".format(epoch));



                if(self.system_dict["dataset"]["label_type"] == "single"):
                    if(val_acc > best_acc):
                        best_acc = val_acc;
                        best_acc_epoch = epoch;
                        best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());
                        torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "best_model");
                        self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                        self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
                else:
                    if(val_loss < best_loss):
                        best_loss = val_loss;
                        best_acc_epoch = epoch;
                        best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());
                        torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "best_model");
                        self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                        self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
                    if(val_acc > best_acc):
                        best_acc = val_acc;


                time_elapsed_since = time.time() - since;
                if("training_time" in self.system_dict["training"]["outputs"].keys()):
                    minutes, seconds = self.system_dict["training"]["outputs"]["training_time"].split(" ");
                    minutes = int(minutes[:len(minutes)-1]);
                    seconds = int(seconds[:len(seconds)-1]);
                    time_elapsed_since += minutes*60 + seconds;
                self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);


                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history), allow_pickle=True);

                    create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                    create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);
                
                torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "resume_state");

                if(self.system_dict["local"]["learning_rate_scheduler"]):
                    if(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] == "reduceonplateaulr"):
                        self.system_dict["local"]["learning_rate_scheduler"].step(epoch_loss);
                    else:
                        self.system_dict["local"]["learning_rate_scheduler"].step();

                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    self.custom_print("");
                    self.custom_print("");

                if(self.system_dict["training"]["settings"]["display_progress"]):
                    for param_group in self.system_dict["local"]["optimizer"].param_groups:
                        curr_lr = param_group['lr'];
                    self.custom_print("    curr_lr - {}".format(curr_lr));
                    self.custom_print('    [Epoch %d] Train-acc: %.3f, Train-loss: %.3f | Val-acc: %3f, Val-loss: %.3f, | time: %.1f sec' %
                                 (epoch+1, train_acc, train_loss, val_acc, val_loss, time.time() - since));
                    self.custom_print("");
                self.system_dict["training"]["outputs"]["epochs_completed"] = epoch+1;
                save(self.system_dict);

            if(self.system_dict["training"]["settings"]["display_progress"]):
                self.custom_print('    Training completed in: {:.0f}m {:.0f}s'.format(time_elapsed_since // 60, time_elapsed_since % 60))
                self.custom_print('    Best val Acc:          {:4f}'.format(best_acc))
                self.custom_print("");
            
            






        
        elif(self.system_dict["states"]["eval_infer"]):
            msg = "Cannot train in testing (eval_infer) mode.\n";
            msg += "Tip - use new_experiment function with a copy_from argument.\n";
            raise ConstraintError(msg);



        else:
            self.custom_print("Training Start");
            self.system_dict = load_optimizer(self.system_dict);
            self.system_dict = load_scheduler(self.system_dict);
            self.system_dict = load_loss(self.system_dict);


            self.system_dict["training"]["status"] = False;

            pid = os.getpid();

            if(self.system_dict["training"]["settings"]["save_training_logs"]):
                val_acc_history = [];
                train_acc_history = [];
                val_loss_history = [];
                train_loss_history = [];

            num_batch_train = len(self.system_dict["local"]["data_loaders"]["train"]);
            num_batch_val = len(self.system_dict["local"]["data_loaders"]["val"]);

            best_acc = 0.0;
            best_loss = 1000.0;
            best_acc_epoch = 0;
            max_gpu_usage = 0;
            best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());

            for epoch in range(self.system_dict["hyper-parameters"]["num_epochs"]):
                if(self.system_dict["training"]["settings"]["display_progress"]):
                    self.custom_print('    Epoch {}/{}'.format(epoch+1, self.system_dict["hyper-parameters"]["num_epochs"]))
                    self.custom_print('    ' + '-' * 10)

                since = time.time();

                for phase in ['train', 'val']:
                    if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                        pbar=tqdm(total=len(self.system_dict["local"]["data_loaders"][phase]));

                    if phase == 'train':
                        self.system_dict["local"]["model"].train() 
                    else:
                        self.system_dict["local"]["model"].eval()  

                    running_loss = 0.0
                    running_corrects = 0
                    total_labels = 0


                    for inputs, labels in self.system_dict["local"]["data_loaders"][phase]:
                        if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                            pbar.update();
                        inputs = inputs.to(self.system_dict["local"]["device"]);
                        labels = labels.to(self.system_dict["local"]["device"]);

                        self.system_dict["local"]["optimizer"].zero_grad();


                        with torch.set_grad_enabled(phase == 'train'):
                            if(self.system_dict["model"]["params"]["model_name"]):
                                if "inception" in self.system_dict["model"]["params"]["model_name"] and phase == 'train':
                                    outputs, aux_outputs = self.system_dict["local"]["model"](inputs)
                                    loss1 = self.system_dict["local"]["criterion"](outputs, labels)
                                    loss2 = self.system_dict["local"]["criterion"](aux_outputs, labels)
                                    loss = loss1 + 0.4*loss2
                                else:
                                    outputs = self.system_dict["local"]["model"](inputs)
                                    loss = self.system_dict["local"]["criterion"](outputs, labels)
                            else:
                                outputs = self.system_dict["local"]["model"](inputs)
                                loss = self.system_dict["local"]["criterion"](outputs, labels)

                            _, preds = torch.max(outputs, 1)

                            if phase == 'train':
                                loss.backward()
                                self.system_dict["local"]["optimizer"].step()


                        running_loss += loss.item() * inputs.size(0)
                        if(self.system_dict["dataset"]["label_type"] == "single"):
                            running_corrects += torch.sum(preds == labels.data)
                        else:
                            labels = labels.cpu().detach().numpy()[0]
                            list_classes = [];
                            list_labels = [];
                            raw_scores = outputs.cpu().detach().numpy()[0];

                            for i in range(len(raw_scores)):
                                prob = logistic.cdf(raw_scores[i])
                                if(prob > 0.5):
                                    list_classes.append(self.system_dict["dataset"]["params"]["classes"][i])

                            for i in range(len(labels)):
                                if(labels[i]):
                                    list_labels.append(self.system_dict["dataset"]["params"]["classes"][i])

                            for i in range(len(list_labels)):
                                actual = list_labels[i];
                                if actual in list_classes:
                                    correct = True;
                                else:
                                    correct = False;

                                index = self.system_dict["dataset"]["params"]["classes"].index(actual);
                                total_labels += 1;

                                if(correct):
                                    running_corrects += 1;


                    epoch_loss = running_loss / len(self.system_dict["local"]["data_loaders"][phase].dataset)
                    
                    if(self.system_dict["dataset"]["label_type"] == "single"):
                        epoch_acc = running_corrects.double() / len(self.system_dict["local"]["data_loaders"][phase].dataset);
                    else:
                        epoch_acc = running_corrects/total_labels;


                    if(not os.getcwd() == "/kaggle/working"):
                        if(self.system_dict["model"]["params"]["use_gpu"]):
                            GPUs = GPUtil.getGPUs()
                            gpuMemoryUsed = GPUs[0].memoryUsed
                            if(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] < int(gpuMemoryUsed)):
                                self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = int(gpuMemoryUsed);
                    else:
                        gpuMemoryUsed = 0;
                        self.system_dict["training"]["outputs"]["max_gpu_memory_usage"] = 0;


                    if(self.system_dict["training"]["settings"]["save_training_logs"]):
                        if(self.system_dict["dataset"]["label_type"] == "single"):
                            if phase == 'val':
                                val_acc = epoch_acc;
                                val_loss = epoch_loss;
                                val_acc_history.append(epoch_acc.cpu().detach().numpy());
                                val_loss_history.append(epoch_loss);
                            else:
                                train_acc = epoch_acc;
                                train_loss = epoch_loss;
                                train_acc_history.append(epoch_acc.cpu().detach().numpy());
                                train_loss_history.append(epoch_loss);
                        else:
                            if phase == 'val':
                                val_acc = epoch_acc;
                                val_loss = epoch_loss;
                                val_acc_history.append(epoch_acc);
                                val_loss_history.append(epoch_loss);
                            else:
                                train_acc = epoch_acc;
                                train_loss = epoch_loss;
                                train_acc_history.append(epoch_acc);
                                train_loss_history.append(epoch_loss);


                if(self.system_dict["training"]["settings"]["save_intermediate_models"]):
                    torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + 
                        self.system_dict["training"]["settings"]["intermediate_model_prefix"] + "{}".format(epoch));


                if(self.system_dict["dataset"]["label_type"] == "single"):
                    if(val_acc > best_acc):
                        best_acc = val_acc;
                        best_acc_epoch = epoch;
                        best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());
                        torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "best_model");
                        self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                        self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
                else:
                    if(val_loss < best_loss):
                        best_loss = val_loss;
                        best_acc_epoch = epoch;
                        best_model_wts = copy.deepcopy(self.system_dict["local"]["model"].state_dict());
                        torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "best_model");
                        self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
                        self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
                    if(val_acc > best_acc):
                        best_acc = val_acc;

                time_elapsed_since = time.time() - since;
                if("training_time" in self.system_dict["training"]["outputs"].keys()):
                    minutes, seconds = self.system_dict["training"]["outputs"]["training_time"].split(" ");
                    minutes = int(minutes[:len(minutes)-1]);
                    seconds = int(seconds[:len(seconds)-1]);
                    time_elapsed_since += minutes*60 + seconds;
                self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);


                if(self.system_dict["training"]["settings"]["save_training_logs"]):
                    np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history), allow_pickle=True);
                    np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history), allow_pickle=True);

                    create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                    create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);
                
                torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "resume_state");

                if(self.system_dict["local"]["learning_rate_scheduler"]):
                    if(self.system_dict["hyper-parameters"]["learning_rate_scheduler"]["name"] == "reduceonplateaulr"):
                        self.system_dict["local"]["learning_rate_scheduler"].step(epoch_loss);
                    else:
                        self.system_dict["local"]["learning_rate_scheduler"].step();

                if(self.system_dict["training"]["settings"]["display_progress_realtime"] and self.system_dict["verbose"]):
                    self.custom_print("");
                    self.custom_print("");

                if(self.system_dict["training"]["settings"]["display_progress"]):
                    for param_group in self.system_dict["local"]["optimizer"].param_groups:
                        curr_lr = param_group['lr'];
                    self.custom_print("    curr_lr - {}".format(curr_lr));
                    self.custom_print('    [Epoch %d] Train-acc: %.3f, Train-loss: %.3f | Val-acc: %3f, Val-loss: %.3f, | time: %.1f sec' %
                                 (epoch+1, train_acc, train_loss, val_acc, val_loss, time.time() - since));
                    self.custom_print("");
                self.system_dict["training"]["outputs"]["epochs_completed"] = epoch+1;
                save(self.system_dict);

            if(self.system_dict["training"]["settings"]["display_progress"]):
                self.custom_print('    Training completed in: {:.0f}m {:.0f}s'.format(time_elapsed_since // 60, time_elapsed_since % 60))
                self.custom_print('    Best val Acc:          {:4f}'.format(best_acc))
                self.custom_print("");


        
        if(not self.system_dict["states"]["eval_infer"]):
            self.custom_print("Training End");
            self.custom_print("");
            self.system_dict["training"]["outputs"]["best_val_acc"] = "{:4f}".format(best_acc);
            self.system_dict["training"]["outputs"]["best_val_acc_epoch_num"] = best_acc_epoch;
            self.system_dict["training"]["outputs"]["training_time"] = "{:.0f}m {:.0f}s".format(time_elapsed_since // 60, time_elapsed_since % 60);
            self.system_dict["training"]["outputs"]["max_gpu_usage"] = str(self.system_dict["training"]["outputs"]["max_gpu_memory_usage"]) + " Mb";

            torch.save(self.system_dict["local"]["model"], self.system_dict["model_dir"] + "final");

            if(self.system_dict["training"]["settings"]["save_training_logs"]): 
                self.custom_print("Training Outputs");
                self.custom_print("    Model Dir:   {}".format(self.system_dict["model_dir"]));
                self.custom_print("    Log Dir:     {}".format(self.system_dict["log_dir"]));
                self.custom_print("    Final model: {}".format("final"));
                self.custom_print("    Best model:  {}".format("best_model"));
                self.custom_print("    Log 1 - Validation accuracy history log: {}".format("val_acc_history.npy"));
                self.custom_print("    Log 2 - Validation loss history log:     {}".format("val_loss_history.npy"));
                self.custom_print("    Log 3 - Training accuracy history log:   {}".format("train_acc_history.npy"));
                self.custom_print("    Log 4 - Training loss history log:       {}".format("train_loss_history.npy"));
                self.custom_print("    Log 5 - Training curve:                  {}".format("train_loss_history.npy"));
                self.custom_print("    Log 6 - Validation curve:                {}".format("train_loss_history.npy"));
                self.custom_print("");

                np.save(self.system_dict["log_dir"] + "val_acc_history.npy", np.array(val_acc_history), allow_pickle=True);
                np.save(self.system_dict["log_dir"] + "val_loss_history.npy", np.array(val_loss_history), allow_pickle=True);
                np.save(self.system_dict["log_dir"] + "train_acc_history.npy", np.array(train_acc_history), allow_pickle=True);
                np.save(self.system_dict["log_dir"] + "train_loss_history.npy", np.array(train_loss_history), allow_pickle=True);
                
                self.system_dict["training"]["outputs"]["log_val_acc_history"] = self.system_dict["log_dir"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history"] = self.system_dict["log_dir"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history"] = self.system_dict["log_dir"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history"] = self.system_dict["log_dir"] + "train_loss_history.npy";

                self.system_dict["training"]["outputs"]["log_val_acc_history_relative"] = self.system_dict["log_dir_relative"] + "val_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_val_loss_history_relative"] = self.system_dict["log_dir_relative"] + "val_loss_history.npy";
                self.system_dict["training"]["outputs"]["log_train_acc_history_relative"] = self.system_dict["log_dir_relative"] + "train_acc_history.npy";
                self.system_dict["training"]["outputs"]["log_train_loss_history_relative"] = self.system_dict["log_dir_relative"] + "train_loss_history.npy";


                create_train_test_plots_accuracy([train_acc_history, val_acc_history], ["Epoch Num", "Accuracy"], self.system_dict["log_dir"], show_img=False, save_img=True);
                create_train_test_plots_loss([train_loss_history, val_loss_history], ["Epoch Num", "Loss"], self.system_dict["log_dir"], show_img=False, save_img=True);

            self.system_dict["training"]["status"] = True;
        

    ###############################################################################################################################################


from monk.pytorch.models.imports import *
from monk.system.imports import *


@accepts(dict, int, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def get_layer(network_layer, num_ftrs):
    '''
    Get layer for appending it to transfer learning base model

    Args:
        network_layer (dict): Dictionary conatining all params relating to the layer
        num_ftrs (int): Number of input features to the layer

    Returns:
        neural network layer
        int: Number of output features from the layer
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];
    
    if(layer_name == "linear"):
        layer = nn.Linear(num_ftrs, layer_params["out_features"])
        num_ftrs = layer_params["out_features"];
    elif(layer_name == "dropout"):
        layer = nn.Dropout(p=layer_params["p"]);
    elif(layer_name == "elu"):
        layer = nn.ELU(alpha=layer_params["alpha"]);
    elif(layer_name == "hardshrink"):
        layer = nn.Hardshrink(lambd=layer_params["lambd"]);
    elif(layer_name == "hardtanh"):
        layer = nn.Hardtanh(min_val=layer_params["min_val"], max_val=layer_params["max_val"]);
    elif(layer_name == "leakyrelu"):
        layer = nn.LeakyReLU(negative_slope=layer_params["negative_slope"]);
    elif(layer_name == "logsigmoid"):
        layer = nn.LogSigmoid();
    elif(layer_name == "prelu"):
        layer = nn.PReLU(num_parameters=layer_params["num_parameters"], init=layer_params["init"]);
    elif(layer_name == "relu"):
        layer = nn.ReLU();
    elif(layer_name == "relu6"):
        layer = nn.ReLU6();
    elif(layer_name == "rrelu"):
        layer = nn.RReLU(lower=layer_params["lower"], upper=layer_params["upper"]);
    elif(layer_name == "selu"):
        layer = nn.SELU();
    elif(layer_name == "celu"):
        layer = nn.CELU(alpha=layer_params["alpha"]);
    elif(layer_name == "sigmoid"):
        layer = nn.Sigmoid();
    elif(layer_name == "softplus"):
        layer = nn.Softplus(beta=layer_params["beta"], threshold=layer_params["threshold"]);
    elif(layer_name == "softshrink"):
        layer = nn.Softshrink(lambd=layer_params["lambd"]);
    elif(layer_name == "softsign"):
        layer = nn.Softsign();
    elif(layer_name == "tanh"):
        layer = nn.Tanh();
    elif(layer_name == "tanhshrink"):
        layer = nn.Tanhshrink();
    elif(layer_name == "threshold"):
        layer = nn.Threshold(threshold=layer_params["threshold"], value=layer_params["value"]);
    elif(layer_name == "softmin"):
        layer = nn.Softmin();
    elif(layer_name == "softmax"):
        layer = nn.Softmax();
    elif(layer_name == "logsoftmax"):
        layer = nn.LogSoftmax();

    return layer, num_ftrs;






@accepts(dict, num_neurons=int, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_linear(system_dict, num_neurons=512, final_layer=False):
    '''
    Append dense (fully connected) layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        num_neurons (int): Number of neurons in the dense layer
        final_layer (bool): If True, then number of neurons are directly set as number of classes in dataset for single label type classification

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "linear";
    tmp["params"] = {};
    tmp["params"]["out_features"] = num_neurons;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, probability=float, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def layer_dropout(system_dict, probability=0.5, final_layer=False):
    '''
    Append dropout layer to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        probability (float): Droping probability of neurons in next layer
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "dropout";
    tmp["params"] = {};
    tmp["params"]["p"] = probability;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, alpha=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_elu(system_dict, alpha=1.0, final_layer=False):
    '''
    Append exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        alpha (float): Multiplicatve factor.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "elu";
    tmp["params"] = {};
    tmp["params"]["alpha"] = alpha;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, lambd=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_hardshrink(system_dict, lambd=0.5, final_layer=False):
    '''
    Append hardshrink activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        lambd (float): thereshold limit value
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "hardshrink";
    tmp["params"] = {};
    tmp["params"]["lambd"] = lambd;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, min_val=[int, float], max_val=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_hardtanh(system_dict, min_val=-1.0, max_val=1.0, final_layer=False):
    '''
    Append hardtanh activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        min_val (float): lower thereshold limit value
        max_val (float): upper thereshold limit value
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "hardtanh";
    tmp["params"] = {};
    tmp["params"]["min_val"] = min_val;
    tmp["params"]["max_val"] = max_val;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, negative_slope=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_leakyrelu(system_dict, negative_slope=0.01, final_layer=False):
    '''
    Append Leaky - ReLU activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        negative_slope (float): Multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "leakyrelu";
    tmp["params"] = {};
    tmp["params"]["negative_slope"] = negative_slope; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_logsigmoid(system_dict, final_layer=False):
    '''
    Append logsigmoid activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "logsigmoid";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, num_parameters=int, init=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_prelu(system_dict, num_parameters=1, init=0.25, final_layer=False):
    '''
    Append Learnable parameterized rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        init (float): Initialization value for multiplicatve factor towards negative spectrum of real numbers.
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "prelu";
    tmp["params"] = {};
    tmp["params"]["num_parameters"] = num_parameters; 
    tmp["params"]["init"] = init; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_relu(system_dict, final_layer=False):
    '''
    Append rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "relu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_relu6(system_dict, final_layer=False):
    '''
    Append rectified linear unit activation - 6 to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "relu6";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, lower=[int, float], upper=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_rrelu(system_dict, lower=0.125, upper=0.333, final_layer=False):
    '''
    Append randomized rectified linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        lower (float): Lower limit of uniform distribution for the ramdomized multiplicative factor
        upper (float): Upper limit of uniform distribution for the ramdomized multiplicative factor
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "rrelu";
    tmp["params"] = {};
    tmp["params"]["lower"] = lower; 
    tmp["params"]["upper"] = upper; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_selu(system_dict, final_layer=False):
    '''
    Append scaled exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "selu";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;


@accepts(dict, alpha=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_celu(system_dict, alpha=1.0, final_layer=False):
    '''
    Append continuously differentiable exponential linear unit activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        alpha (float): Multiplicative factor
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "celu";
    tmp["params"] = {};
    tmp["params"]["alpha"] = alpha;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_sigmoid(system_dict, final_layer=False):
    '''
    Append sigmoid activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "sigmoid";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, beta=[int, float], threshold=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softplus(system_dict, beta=1, threshold=20, final_layer=False):
    '''
    Append softplus activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        beta (int): Multiplicative factor
        threshold (int): softplus (thresholded relu) limit 
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softplus";
    tmp["params"] = {};
    tmp["params"]["beta"] = beta; 
    tmp["params"]["threshold"] = threshold; 
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, lambd=[int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softshrink(system_dict, lambd=0.5, final_layer=False):
    '''
    Append softshrink activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        lambd (float): thereshold limit value
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softshrink";
    tmp["params"] = {};
    tmp["params"]["lambd"] = lambd;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softsign(system_dict, final_layer=False):
    '''
    Append softsign activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softsign";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_tanh(system_dict, final_layer=False):
    '''
    Append tanh activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "tanh";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_tanhshrink(system_dict, final_layer=False):
    '''
    Append tanhshrink activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "tanhshrink";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, [int, float], [int, float], final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_threshold(system_dict, threshold, value, final_layer=False):
    '''
    Append threshold activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        threshold (float): thereshold limit value
        value (float): replacement value if input is off-limits
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "threshold";
    tmp["params"] = {};
    tmp["params"]["value"] = value; 
    tmp["params"]["threshold"] = threshold;
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;




@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softmin(system_dict, final_layer=False):
    '''
    Append softmin activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softmin";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_softmax(system_dict, final_layer=False):
    '''
    Append softmax activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "softmax";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, final_layer=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def activation_logsoftmax(system_dict, final_layer=False):
    '''
    Append log-softmax activation to base network in transfer learning

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        final_layer (bool): Indicator that this layer marks the end of network.

    Returns:
        dict: updated system dict
    '''
    tmp = {};
    tmp["name"] = "logsoftmax";
    tmp["params"] = {};
    system_dict["model"]["custom_network"].append(tmp);
    system_dict["model"]["final_layer"] = final_layer;

    return system_dict;



@accepts(dict, [int, tuple], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=True)
def custom_model_get_layer(network_layer, current_in_shape):
    '''
    Get layer for custom network design

    Args:
        network_layer (dict): Network layer as dict
        current_in_shape (tuple): Input shape to current layer 

    Returns:
        neural network layer: Actual layer in native backend library
        tuple: Output shape from the current layer
    '''
    layer_name = network_layer["name"];
    layer_params = network_layer["params"];

    if(layer_name == "convolution1d"):
        layer, current_in_shape = custom_model_layer_convolution1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "convolution2d"):
        layer, current_in_shape = custom_model_layer_convolution2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "convolution3d"):
        layer, current_in_shape = custom_model_layer_convolution3d(layer_params, current_in_shape);
        return layer, current_in_shape;

    elif(layer_name == "transposed_convolution1d"):
        layer, current_in_shape = custom_model_layer_transposed_convolution1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "transposed_convolution2d"):
        layer, current_in_shape = custom_model_layer_transposed_convolution2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "transposed_convolution3d"):
        layer, current_in_shape = custom_model_layer_transposed_convolution3d(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "max_pooling1d"):
        layer, current_in_shape = custom_model_layer_max_pooling1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "max_pooling2d"):
        layer, current_in_shape = custom_model_layer_max_pooling2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "max_pooling3d"):
        layer, current_in_shape = custom_model_layer_max_pooling3d(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "average_pooling1d"):
        layer, current_in_shape = custom_model_layer_average_pooling1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "average_pooling2d"):
        layer, current_in_shape = custom_model_layer_average_pooling2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "average_pooling3d"):
        layer, current_in_shape = custom_model_layer_average_pooling3d(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "global_max_pooling1d"):
        layer, current_in_shape = custom_model_layer_global_max_pooling1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "global_max_pooling2d"):
        layer, current_in_shape = custom_model_layer_global_max_pooling2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "global_max_pooling3d"):
        layer, current_in_shape = custom_model_layer_global_max_pooling3d(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "global_average_pooling1d"):
        layer, current_in_shape = custom_model_layer_global_average_pooling1d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "global_average_pooling2d"):
        layer, current_in_shape = custom_model_layer_global_average_pooling2d(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "global_average_pooling3d"):
        layer, current_in_shape = custom_model_layer_global_average_pooling3d(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "fully_connected"):
        layer, current_in_shape = custom_model_layer_fully_connected(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "flatten"):
        layer, current_in_shape = custom_model_layer_flatten(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "dropout"):
        layer, current_in_shape = custom_model_layer_dropout(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "identity"):
        layer, current_in_shape = custom_model_layer_identity(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "batch_normalization"):
        layer, current_in_shape = custom_model_layer_batch_normalization(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "instance_normalization"):
        layer, current_in_shape = custom_model_layer_instance_normalization(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "layer_normalization"):
        layer, current_in_shape = custom_model_layer_layer_normalization(layer_params, current_in_shape);
        return layer, current_in_shape;


    elif(layer_name == "relu"):
        layer, current_in_shape = custom_model_activation_relu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "sigmoid"):
        layer, current_in_shape = custom_model_activation_sigmoid(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "tanh"):
        layer, current_in_shape = custom_model_activation_tanh(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "softplus"):
        layer, current_in_shape = custom_model_activation_softplus(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "softsign"):
        layer, current_in_shape = custom_model_activation_softsign(layer_params, current_in_shape);
        return layer, current_in_shape;

    elif(layer_name == "elu"):
        layer, current_in_shape = custom_model_activation_elu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "leaky_relu"):
        layer, current_in_shape = custom_model_activation_leaky_relu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "prelu"):
        layer, current_in_shape = custom_model_activation_prelu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "selu"):
        layer, current_in_shape = custom_model_activation_selu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "hardshrink"):
        layer, current_in_shape = custom_model_activation_hardshrink(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "hardtanh"):
        layer, current_in_shape = custom_model_activation_hardtanh(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "logsigmoid"):
        layer, current_in_shape = custom_model_activation_logsigmoid(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "relu6"):
        layer, current_in_shape = custom_model_activation_relu6(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "rrelu"):
        layer, current_in_shape = custom_model_activation_rrelu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "celu"):
        layer, current_in_shape = custom_model_activation_celu(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "softshrink"):
        layer, current_in_shape = custom_model_activation_softshrink(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "tanhshrink"):
        layer, current_in_shape = custom_model_activation_tanhshrink(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "threshold"):
        layer, current_in_shape = custom_model_activation_threshold(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "softmin"):
        layer, current_in_shape = custom_model_activation_softmin(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "softmax"):
        layer, current_in_shape = custom_model_activation_softmax(layer_params, current_in_shape);
        return layer, current_in_shape;
    elif(layer_name == "logsoftmax"):
        layer, current_in_shape = custom_model_activation_logsoftmax(layer_params, current_in_shape);
        return layer, current_in_shape;





@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution1d(params, current_in_shape):
    '''
    Append 1d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];

    layer = nn.Conv1d(in_channels,
                        params["output_channels"], 
                        params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        bias=params["use_bias"]);

    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])
    
    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution2d(params, current_in_shape):
    '''
    Append 2d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];

    layer = nn.Conv2d(in_channels,
                        params["output_channels"], 
                        params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        bias=params["use_bias"]);

    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])


    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_convolution3d(params, current_in_shape):
    '''
    Append 3d-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["dilation"]*(params["kernel_size"] - 1) - params["stride"] + 1)//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];
        
    layer = nn.Conv3d(in_channels,
                        params["output_channels"], 
                        params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"], 
                        dilation=params["dilation"], 
                        groups=params["groups"], 
                        bias=params["use_bias"]);

    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])
    
    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution1d(params, current_in_shape):
    '''
    Append 1d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];

    layer = nn.ConvTranspose1d(in_channels,
                                params["output_channels"], 
                                params["kernel_size"], 
                                stride=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"],
                                output_padding=params["output_padding"], 
                                bias=params["use_bias"])

    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution2d(params, current_in_shape):
    '''
    Append 2d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];

    layer = nn.ConvTranspose2d(in_channels,
                                params["output_channels"], 
                                params["kernel_size"], 
                                stride=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"],
                                output_padding=params["output_padding"], 
                                bias=params["use_bias"])

    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])

    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_transposed_convolution3d(params, current_in_shape):
    '''
    Append 3d-transposed-convolution to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-transposed-convolution layer
    '''
    if(params["padding"] == "in_eq_out" and params["stride"]==1):
        params["padding"] = (params["kernel_size"] + params["output_padding"])//2;
    elif(params["padding"] == "in_eq_out" and params["stride"]!=1):
        params["padding"] = 0;

    in_channels = current_in_shape[0];

    layer = nn.ConvTranspose3d(in_channels,
                                params["output_channels"], 
                                params["kernel_size"], 
                                stride=params["stride"], 
                                padding=params["padding"], 
                                dilation=params["dilation"], 
                                groups=params["groups"],
                                output_padding=params["output_padding"], 
                                bias=params["use_bias"])

    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])

    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling1d(params, current_in_shape):
    '''
    Append 1d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-max-pooling layer
    '''

    in_channels = current_in_shape[0];

    layer = nn.MaxPool1d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        return_indices=params["return_indices"]);
    
    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling2d(params, current_in_shape):
    '''
    Append 2d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-max-pooling layer
    '''

    in_channels = current_in_shape[0];

    layer = nn.MaxPool2d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        return_indices=params["return_indices"]);
    
    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_max_pooling3d(params, current_in_shape):
    '''
    Append 3d-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-max-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.MaxPool3d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"],
                        ceil_mode=params["ceil_mode"],
                        return_indices=params["return_indices"]);
    
    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])

    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling1d(params, current_in_shape):
    '''
    Append 1d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-average-pooling layer
    '''

    in_channels = current_in_shape[0];

    layer = nn.AvgPool1d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"]);
    
    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling2d(params, current_in_shape):
    '''
    Append 2d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-average-pooling layer
    '''

    in_channels = current_in_shape[0];

    layer = nn.AvgPool2d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"]);
    
    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_average_pooling3d(params, current_in_shape):
    '''
    Append 3d-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-average-pooling layer
    '''

    in_channels = current_in_shape[0];

    layer = nn.AvgPool3d(params["kernel_size"], 
                        stride=params["stride"], 
                        padding=params["padding"]);
    
    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling1d(params, current_in_shape):
    '''
    Append 1d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-max-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveMaxPool1d(output_size=1);
    
    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling2d(params, current_in_shape):
    '''
    Append 2d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-max-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveMaxPool2d(output_size=1);
    
    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_max_pooling3d(params, current_in_shape):
    '''
    Append 3d-global-max-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-max-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveMaxPool3d(output_size=1);
    
    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling1d(params, current_in_shape):
    '''
    Append 1d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 1d-global-average-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveAvgPool1d(output_size=1);
    
    c, w = current_in_shape
    x = torch.randn(1, c, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling2d(params, current_in_shape):
    '''
    Append 2d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 2d-global-average-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveAvgPool2d(output_size=1);
    
    c, h, w = current_in_shape
    x = torch.randn(1, c, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3])

    return layer, current_in_shape


@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_global_average_pooling3d(params, current_in_shape):
    '''
    Append 3d-global-average-pooling to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: 3d-global-average-pooling layer
    '''
    in_channels = current_in_shape[0];

    layer = nn.AdaptiveAvgPool3d(output_size=1);
    
    c, d, h, w = current_in_shape
    x = torch.randn(1, c, d, h, w);
    y = layer(x)
    current_in_shape = (y.shape[1], y.shape[2], y.shape[3], y.shape[4])

    return layer, current_in_shape



@accepts(dict, tuple, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_flatten(params, current_in_shape):
    '''
    Append flatten to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: flatten layer
    '''
    
    in_channels = current_in_shape[0];

    layer = nn.Flatten();

    if(len(current_in_shape) == 2):
        c, w = current_in_shape
        x = torch.randn(1, c, w);
        y = layer(x)
        current_in_shape = (y.shape[1])
    elif(len(current_in_shape) == 3):
        c, h, w = current_in_shape
        x = torch.randn(1, c, h, w);
        y = layer(x)
        current_in_shape = (y.shape[1])
    else:
        c, d, h, w = current_in_shape
        x = torch.randn(1, c, d, h, w);
        y = layer(x)
        current_in_shape = (y.shape[1])

    return layer, current_in_shape;


@accepts(dict, [int, tuple], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_fully_connected(params, current_in_shape):
    '''
    Append fc (dense) to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: fc (dense) layer
    '''
    if(type(current_in_shape) == int):
        in_feat = current_in_shape;
    elif(type(current_in_shape) == tuple):
        in_feat = current_in_shape[0];

    layer = nn.Linear(in_feat, 
                    params["units"],
                    bias=params["use_bias"]);

    x = torch.randn(1, in_feat);
    y = layer(x)
    current_in_shape = (y.shape[1])

    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_dropout(params, current_in_shape):
    '''
    Append dropout to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: dropout layer
    '''
    layer = nn.Dropout(p=params["drop_probability"]);
    
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_identity(params, current_in_shape):
    '''
    Append idenity to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: idenity layer
    '''
    layer = nn.Identity();
    
    return layer, current_in_shape;


@accepts(dict, [tuple, int],  post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_batch_normalization(params, current_in_shape):
    '''
    Append batchnorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural netwrk layer: batchnorm layer
    '''
    in_channels = current_in_shape[0];

    if(len(current_in_shape) == 2):
        layer = nn.BatchNorm1d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])

    elif(len(current_in_shape) == 3):
        layer = nn.BatchNorm2d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])

    elif(len(current_in_shape) == 4):
        layer = nn.BatchNorm3d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])


    return layer, current_in_shape;



@accepts(dict, [tuple, int],  post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_instance_normalization(params, current_in_shape):
    '''
    Append instancenorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: instancenorm layer
    '''
    in_channels = current_in_shape[0];

    if(len(current_in_shape) == 2):
        layer = nn.InstanceNorm1d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])

    elif(len(current_in_shape) == 3):
        layer = nn.InstanceNorm2d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])

    elif(len(current_in_shape) == 4):
        layer = nn.InstanceNorm3d(in_channels, 
                        eps=params["epsilon"], 
                        momentum=params["moving_average_momentum"], 
                        affine=params["use_trainable_parameters"], 
                        track_running_stats=params["use_trainable_parameters"])


    return layer, current_in_shape;



@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_layer_layer_normalization(params, current_in_shape):
    '''
    Append layernorm to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network layer: layernorm layer
    '''
    layer = nn.LayerNorm(list(current_in_shape), 
                        eps=params["epsilon"], 
                        elementwise_affine=params["use_trainable_parameters"]);
    return layer, current_in_shape;



@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_relu(params, current_in_shape):
    '''
    Append relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: relu activation
    '''
    layer = nn.ReLU();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_sigmoid(params, current_in_shape):
    '''
    Append sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: sigmoid activation
    '''
    layer = nn.Sigmoid();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_tanh(params, current_in_shape):
    '''
    Append tanh to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: tanh activation
    '''
    layer = nn.Tanh();
    return layer, current_in_shape;

@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softplus(params, current_in_shape):
    '''
    Append softplus to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softplus activation
    '''
    layer = nn.Softplus(beta=params["beta"], 
                            threshold=params["threshold"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softsign(params, current_in_shape):
    '''
    Append softsign to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softsign activation
    '''
    layer = nn.Softsign();
    return layer, current_in_shape;



@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_elu(params, current_in_shape):
    '''
    Append elu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: elu activation
    '''
    layer = nn.ELU(alpha=params["alpha"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_leaky_relu(params, current_in_shape):
    '''
    Append leaky-relu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: leaky-relu activation
    '''
    layer = nn.LeakyReLU(negative_slope=params["alpha"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_prelu(params, current_in_shape):
    '''
    Append prelu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: prelu activation
    '''
    layer = nn.PReLU();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_selu(params, current_in_shape):
    '''
    Append selu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: selu activation
    '''
    layer = nn.SELU();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_hardshrink(params, current_in_shape):
    '''
    Append hardshrink to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: hardshrink activation
    '''
    layer = nn.Hardshrink(lambd=params["threshold"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_hardtanh(params, current_in_shape):
    '''
    Append hardtanh to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: hardtanh activation
    '''
    layer = nn.Hardtanh(min_val=params["min_val"],
                            max_val=params["max_val"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_logsigmoid(params, current_in_shape):
    '''
    Append log-sigmoid to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: log-sigmoid activation
    '''
    layer = nn.LogSigmoid();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_relu6(params, current_in_shape):
    '''
    Append relu6 to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: relu6 activation
    '''

    layer = nn.ReLU6();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_rrelu(params, current_in_shape):
    '''
    Append rrelu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: rrelu activation
    '''
    layer = nn.RReLU();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_celu(params, current_in_shape):
    '''
    Append celu to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: celu activation
    '''
    layer = nn.CELU(alpha=params["alpha"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softshrink(params, current_in_shape):
    '''
    Append softshrink to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softshrink activation
    '''
    layer = nn.Softshrink(lambd=params["threshold"]);
    return layer, current_in_shape;



@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_tanhshrink(params, current_in_shape):
    '''
    Append tanhshrink to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: tanhshrink activation
    '''
    layer = nn.Tanhshrink();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_threshold(params, current_in_shape):
    '''
    Append threshold to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: threshold activation
    '''
    layer = nn.Threshold(params["threshold"], 
                            params["value"]);
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softmin(params, current_in_shape):
    '''
    Append softmin to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softmin activation
    '''
    layer = nn.Softmin();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_softmax(params, current_in_shape):
    '''
    Append softmax to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: softmax activation
    '''
    layer = nn.Softmax();
    return layer, current_in_shape;


@accepts(dict, [tuple, int], post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def custom_model_activation_logsoftmax(params, current_in_shape):
    '''
    Append log-softmax to custom network

    Args:
        params (dict): All layer parameters

    Returns:
        neural network activation: log-softmax activation
    '''
    layer = nn.LogSoftmax();
    return layer, current_in_shape;


class Net_Add(nn.Module):
    '''
    Class for creating a custom block (layer) representing elementwise addition

    Args:
        branches (list): List of layers to merge with elementwise addition
    '''
    def __init__(self, branches):
        super().__init__()
        self.child_names = [];
        
        for i in range(len(branches)):
            vars(self)["body" + str(i)] = nn.Sequential();
            for j in range(len(branches[i])):
                vars(self)["body" + str(i)].add_module("br_{}_{}".format(i, j), branches[i][j]);
            self.child_names.append("body" + str(i));
                
              
        for i, child in enumerate(self.child_names):
            setattr(self, 'body{0}'.format(i), vars(self)["body" + str(i)])
        
        
        
    def forward(self, x):
        for i in range(len(self.child_names)):
            br = getattr(self, 'body{0}'.format(i));
            if(i==0):
                y = br(x);
            else:
                y = y + br(x);
        return y



class Net_Concat(nn.Module):
    '''
    Class for creating a custom block (layer) representing axiswise concatenation

    Args:
        branches (list): List of layers to merge with axiswise concatenation
    '''
    def __init__(self, branches):
        super().__init__()
        self.child_names = [];
        
        for i in range(len(branches)):
            vars(self)["body" + str(i)] = nn.Sequential();
            for j in range(len(branches[i])):
                vars(self)["body" + str(i)].add_module("br_{}_{}".format(i, j), branches[i][j]);
            self.child_names.append("body" + str(i));
                
              
        for i, child in enumerate(self.child_names):
            setattr(self, 'body{0}'.format(i), vars(self)["body" + str(i)])
        
        
        
    def forward(self, x):
        outs = [];
        for i in range(len(self.child_names)):
            br = getattr(self, 'body{0}'.format(i));
            outs.append(br(x));
        return torch.cat(tuple(outs), 1)


from monk.tf_keras_1.testing.imports import *
from monk.system.imports import *

@accepts(str, bool, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_single(img_name, return_raw, system_dict):
    '''
    Run inference on a single image

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
                            

    Returns:
        str: predicted class
        float: prediction score
    '''
    input_size = system_dict["dataset"]["params"]["input_size"];
    normalize = system_dict["local"]["normalize"];
    mean_subtract = system_dict["local"]["mean_subtract"];
    mean = system_dict["local"]["transforms_test"]["mean"];
    std = system_dict["local"]["transforms_test"]["std"];

    img = cv2.imread(img_name, 1);
    img = cv2.resize(img, (input_size, input_size))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB);
    img = img.astype(np.float32);

    if(normalize):
        img = img-mean;
        img = img/std;
    elif(mean_subtract):
        img = img-mean;
    img = np.expand_dims(img, axis=0)



    output = system_dict["local"]["model"].predict(img);

    if(system_dict["dataset"]["params"]["classes"]):
        class_names = list(system_dict["dataset"]["params"]["classes"].keys());
        prediction = class_names[np.argmax(output)];
    else:
        prediction = np.argmax(output);

    score = output[0][np.argmax(output)];

    if(return_raw):
        return prediction, score, output[0];
    else:
        return prediction, score, "";






@accepts(str, bool, float, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def process_multi(img_name, return_raw, img_thresh, system_dict):
    '''
    Run inference on a single image when label type is multi-label

    Args:
        img_name (str): path to image
        return_raw (bool): If True, then output dictionary contains image probability for every class in the set.
                            Else, only the most probable class score is returned back.
        img_thresh (float): Thresholding for multi label image classification.
                            

    Returns:
        list: list of predicted classes
        list: list of predicted scores
    '''

    input_size = system_dict["dataset"]["params"]["input_size"];
    normalize = system_dict["local"]["normalize"];
    mean_subtract = system_dict["local"]["mean_subtract"];
    mean = system_dict["local"]["transforms_test"]["mean"];
    std = system_dict["local"]["transforms_test"]["std"];

    img = cv2.imread(img_name, 1);
    img = cv2.resize(img, (input_size, input_size))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB);
    img = img.astype(np.float32);

    if(normalize):
        img = img-mean;
        img = img/std;
    elif(mean_subtract):
        img = img-mean;
    img = np.expand_dims(img, axis=0)



    outputs = system_dict["local"]["model"].predict(img);

    list_classes = [];
    raw_scores = outputs[0];
    list_scores = [];

    for i in range(len(raw_scores)):
        prob = logistic.cdf(raw_scores[i])
        if(prob > img_thresh):
            list_classes.append(list(system_dict["dataset"]["params"]["classes"].keys())[i])
            list_scores.append(prob)


    if(return_raw):
        return list_classes, list_scores, raw_scores;
    else:
        return list_classes, list_scores, "";





from monk.tf_keras_1.optimizers.imports import *
from monk.system.imports import *

@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def load_optimizer(system_dict):
    '''
    Load Optimizers in training states

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables

    Returns:
        dict: updated system dict
    '''
    optimizer = system_dict["local"]["optimizer"];
    learning_rate = system_dict["hyper-parameters"]["learning_rate"];

    if(optimizer == "sgd"):
        system_dict["local"]["optimizer"] = kro.SGD(
                lr=learning_rate, 
                momentum=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
                nesterov=False,
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);

    elif(optimizer == "nesterov_sgd"):
        system_dict["local"]["optimizer"] = kro.SGD(
                lr=learning_rate, 
                momentum=system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
                nesterov=True,
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);


    elif(optimizer == "rmsprop"):
        system_dict["local"]["optimizer"] = kro.RMSprop(
                lr=learning_rate, 
                rho=system_dict["hyper-parameters"]["optimizer"]["params"]["decay_rate"], 
                epsilon=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"],
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);


    elif(optimizer == "adam"):
        system_dict["local"]["optimizer"] = kro.Adam(
                lr=learning_rate, 
                beta_1=system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], 
                beta_2=system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"], 
                epsilon=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"], 
                amsgrad=system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"],
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);


    elif(optimizer == "nadam"):
        system_dict["local"]["optimizer"] = kro.Nadam(
            lr=learning_rate,
            beta_1=system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], 
            beta_2=system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"], 
            epsilon=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"],
            clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
            clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]
            );

    elif(optimizer == "adamax"):
        system_dict["local"]["optimizer"] = kro.Adamax(
                lr=learning_rate, 
                beta_1=system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"], 
                beta_2=system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"], 
                epsilon=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"],
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);

    elif(optimizer == "adadelta"):
        system_dict["local"]["optimizer"] = kro.Adadelta(
                lr=learning_rate, 
                rho=system_dict["hyper-parameters"]["optimizer"]["params"]["rho"], 
                epsilon=system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"], 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"],
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);


    elif(optimizer == "adagrad"):
        system_dict["local"]["optimizer"] = kro.Adagrad(
                lr=learning_rate, 
                decay=system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"],
                clipnorm=system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"],
                clipvalue=system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"]);




    


    return system_dict;


from monk.pytorch.models.imports import *
from monk.system.imports import *
from monk.pytorch.models.layers import get_layer


@accepts("self", bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def set_parameter_requires_grad(finetune_net, freeze_base_network):
    '''
    Freeze based network as per params set

    Args:
        finetune_net (network): Model network
        freeze_base_network (bool): If True, all trainable params are freezed

    Returns:
        network: Updated Model network
    '''
    if freeze_base_network:
        for param in finetune_net.parameters():
            param.requires_grad = False
    else:
        for param in finetune_net.parameters():
            param.requires_grad = True
    return finetune_net


@accepts(list, int, int, post_trace=False)
#@TraceFunction(trace_args=True, trace_rv=False)
def set_final_layer(custom_network, num_ftrs, num_classes):
    '''
    Setup final sub-network 

    Args:
        custom_network (list): List of dicts containing details on appeded layers to base netwoek in transfer learning
        num_ftrs (int): Number of features coming from base network's last layers
        num_classes (int): Number of classes in the dataset

    Returns:
        layer: Sequential sub-network with added layers 
    '''
    modules = [];
    for i in range(len(custom_network)):
        layer, num_ftrs = get_layer(custom_network[i], num_ftrs);
        modules.append(layer);
    sequential = nn.Sequential(*modules)
    return sequential;


@accepts("self", list, int, set=int, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def create_final_layer(finetune_net, custom_network, num_classes, set=1):
    '''
    Create final sub-network 

    Args:
        finetune_net (network): Initial base network
        custom_network (list):  List of dicts containing details on appeded layers to base netwoek in transfer learning
        num_classes (int): Number of classes in the dataset
        set (int): Select the right set to find the details of outermost layer

    Returns:
        network: Updated base network with appended custom additions
    '''
    if(set == 1):
        num_ftrs = finetune_net.classifier[6].in_features;
        finetune_net.classifier = set_final_layer(custom_network, num_ftrs, num_classes);
    elif(set == 2):
        num_ftrs = finetune_net.classifier.in_features;
        finetune_net.classifier = set_final_layer(custom_network, num_ftrs, num_classes);
    elif(set == 3):
        num_ftrs = finetune_net.fc.in_features;
        finetune_net.fc = set_final_layer(custom_network, num_ftrs, num_classes);
    elif(set == 4):
        num_ftrs = finetune_net.classifier[1].in_features;
        finetune_net.classifier = set_final_layer(custom_network, num_ftrs, num_classes);
    
    return finetune_net;



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def model_to_device(system_dict):
    '''
    Load model weights on device - cpu or gpu 

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    if(system_dict["model"]["params"]["use_gpu"]):
        system_dict["local"]["device"] = torch.device("cuda:0" if torch.cuda.is_available() else "cpu");
        if(torch.cuda.is_available()):
            use_gpu = True;
            system_dict["model"]["params"]["use_gpu"] = use_gpu;
        else:
            use_gpu = False;
            system_dict["model"]["params"]["use_gpu"] = use_gpu;
    else:
        system_dict["local"]["device"] = torch.device("cpu");
    
    system_dict["local"]["model"] = system_dict["local"]["model"].to(system_dict["local"]["device"]);
    return system_dict;


@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def print_grad_stats(system_dict):
    '''
    Print details on which layers are trainable

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        None 
    '''
    print("Model - Gradient Statistics");
    i = 1;
    for name, param in system_dict["local"]["model"].named_parameters():
        if(i%2 != 0):
            print("    {}. {} Trainable - {}".format(i//2+1, name, param.requires_grad ));
        i += 1;
    print("");



@accepts(dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_num_layers(system_dict):
    '''
    Get number of potentially trainable layers

    Args:
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    num_layers = 0;
    layer_names = [];
    for param in system_dict["local"]["model"].named_parameters():
        lname = ".".join(param[0].split(".")[:-1]);
        if lname not in layer_names:
            layer_names.append(lname)

    num_layers = len(layer_names);
    system_dict["model"]["params"]["num_layers"] = num_layers;
    return system_dict;



@accepts(int, dict, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def freeze_layers(num, system_dict):
    '''
    Main function responsible to freeze layers in network

    Args:
        num (int): Number of layers to freeze
        system_dict (dict): System dict containing system state and parameters

    Returns:
        dict: Updated system dict 
    '''
    system_dict = get_num_layers(system_dict);
    num_layers_in_model = system_dict["model"]["params"]["num_layers"];
    if(num > num_layers_in_model):
        msg = "Parameter num > num_layers_in_model\n";
        msg += "Freezing entire network\n";
        msg += "TIP: Total layers: {}".format(num_layers_in_model);
        raise ConstraintError(msg)

    current_num = 0;
    value = False;

    layer_names = [];
    for name,param in system_dict["local"]["model"].named_parameters():
        param.requires_grad = value;
        lname = ".".join(name.split(".")[:-1]);
        if lname not in layer_names:
            layer_names.append(lname);
            current_num += 1;
        if(current_num == num):
            value = True;

    system_dict["local"]["params_to_update"] = []
    layer_names = [];
    for name, param in system_dict["local"]["model"].named_parameters():
        if param.requires_grad == True:
            system_dict["local"]["params_to_update"].append(param);
            lname = ".".join(name.split(".")[:-1]);
            if lname not in layer_names:
                layer_names.append(lname);
    system_dict["model"]["params"]["num_params_to_update"] = len(layer_names);
    system_dict["model"]["status"] = True;

    return system_dict;


@accepts(dict, list, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def get_layer_uid(network_stack, count):
    '''
    Get a unique name for layer in custom network development

    Args:
        network_stack (list): List of list containing custom network details
        count (dict): a unique dictionary mapping number of every type of layer in the network
        system_dict (dict): System dict containing system state and parameters

    Returns:
        str: layer unique name
        dict: updated layer type mapper count
    '''
    if network_stack["uid"]:
        return network_stack["uid"], count;
    else:
        index = layer_names.index(network_stack["name"]);
        network_name = names[index] + str(count[index]);
        count[index] += 1;
        return network_name, count;


from monk.gluon.optimizers.imports import *
from monk.system.imports import *



@accepts(dict, [int, float], momentum=[int, float], momentum_dampening_rate=[int, float], weight_decay=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def sgd(system_dict, learning_rate, momentum=0, weight_decay=0, momentum_dampening_rate=0, clipnorm=0.0, clipvalue=0.0):
    '''
    Select stochastic gradient descent optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        momentum (float): Momentum value for driving the weights towards minima
        weight_decay (float): Value for regularizing weights post every update
        momentum_dampening_rate (float): Reduction rate for momentum
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "sgd";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "sgd";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"] = momentum;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum_dampening_rate"] = momentum_dampening_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;


@accepts(dict, [int, float], momentum=[int, float], momentum_dampening_rate=[int, float], weight_decay=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def nesterov_sgd(system_dict, learning_rate, momentum=0, weight_decay=0, momentum_dampening_rate=0, clipnorm=0.0, clipvalue=0.0):
    '''
    Select stochastic gradient descent optimizer with nesterov acceleration

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        momentum (float): Momentum value for driving the weights towards minima
        weight_decay (float): Value for regularizing weights post every update
        momentum_dampening_rate (float): Reduction rate for momentum
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "nesterov_sgd";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "nesterov_sgd";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"] = momentum;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum_dampening_rate"] = momentum_dampening_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], decay_rate=[int, float], epsilon=[int, float], weight_decay=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def rmsprop(system_dict, learning_rate, decay_rate=0.99, epsilon=1e-08, weight_decay=0, 
    clipnorm=0.0, clipvalue=0.0):
    '''
    Select root mean score prop optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        decay_rate (float): A decay factor of moving average over past squared gradient.
        epsilon (float): A value to avoid division by zero
        weight_decay (float): Value for regularizing weights post every update
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "rmsprop";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "rmsprop";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["decay_rate"] = decay_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], decay_rate=[int, float], epsilon=[int, float], weight_decay=[int, float], 
    momentum=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def momentum_rmsprop(system_dict, learning_rate, decay_rate=0.99, epsilon=1e-08, weight_decay=0, 
    momentum=0.9):
    '''
    Select root mean score prop optimizer with momentum

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        decay_rate (float): A decay factor of moving average over past squared gradient.
        epsilon (float): A value to avoid division by zero
        weight_decay (float): Value for regularizing weights post every update
        momentum (float): Momentum value for driving the weights towards minima
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "rmsprop";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "rmsprop";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["decay_rate"] = decay_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"] = momentum;
    return system_dict;



@accepts(dict, [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], amsgrad=bool, 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def adam(system_dict, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0, amsgrad=False, clipnorm=0.0, clipvalue=0.0):
    '''
    Select ADAM optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        beta1 (float): Exponential decay rate for first momentum estimates
        beta2 (float): Exponential decay rate for first second estimates
        weight_decay (float): Value for regularizing weights post every update
        amsgrad (bool): If True, AMSGrad variant of this algorithm is used
        epsilon (float): A value to avoid division by zero
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "adam";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "adam";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"] = beta1;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"] = beta2;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"] = amsgrad;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;


@accepts(dict, [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def adamax(system_dict, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0, clipnorm=0.0, clipvalue=0.0):
    '''
    Select Adamax optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        beta1 (float): Exponential decay rate for first momentum estimates
        beta2 (float): Exponential decay rate for first second estimates
        weight_decay (float): Value for regularizing weights post every update
        epsilon (float): A value to avoid division by zero
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "adamax";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "adamax";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"] = beta1;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"] = beta2;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], rho=[int, float], epsilon=[int, float], weight_decay=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def adadelta(system_dict, learning_rate, rho=0.9, epsilon=1e-06, weight_decay=0, clipnorm=0.0, clipvalue=0.0):
    '''
    Select Adadelta optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        rho (float): Exponential decay rate for momentum estimates
        weight_decay (float): Value for regularizing weights post every update
        epsilon (float): A value to avoid division by zero
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "adadelta";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "adadelta";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["rho"] = rho;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], learning_rate_decay=[int, float], weight_decay=[int, float], epsilon=[int, float], 
    clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def adagrad(system_dict, learning_rate, learning_rate_decay=0, weight_decay=0, epsilon=0, clipnorm=0.0, clipvalue=0.0):
    '''
    Select Adagrad optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        learning_rate_decay (float): Learning rate decay factor
        weight_decay (float): Value for regularizing weights post every update
        epsilon (float): A value to avoid division by zero
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "adagrad";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "adagrad";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr_decay"] = learning_rate_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], beta1=[int, float], beta2=[int, float], epsilon=[int, float], weight_decay=[int, float], amsgrad=bool, 
    momentum_decay=[int, float], clipnorm=[int, float], clipvalue=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def nesterov_adam(system_dict, learning_rate, beta1=0.9, beta2=0.999, epsilon=1e-08, weight_decay=0, amsgrad=False,
    momentum_decay=0.004, clipnorm=0.0, clipvalue=0.0):
    '''
    Select ADAM optimizer with nesterov momentum acceleration

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        beta1 (float): Exponential decay rate for first momentum estimates
        beta2 (float): Exponential decay rate for first second estimates
        weight_decay (float): Value for regularizing weights post every update
        amsgrad (bool): If True, AMSGrad variant of this algorithm is used
        epsilon (float): A value to avoid division by zero
        clipnorm (float): Gradient clipping factor
        clipvalue (float): Value for clipping

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "nadam";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "nadam";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta1"] = beta1;
    system_dict["hyper-parameters"]["optimizer"]["params"]["beta2"] = beta2;
    system_dict["hyper-parameters"]["optimizer"]["params"]["epsilon"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"] = amsgrad;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum_decay"] = momentum_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipnorm"] = clipnorm;
    system_dict["hyper-parameters"]["optimizer"]["params"]["clipvalue"] = clipvalue;
    return system_dict;



@accepts(dict, [int, float], momentum=[int, float], weight_decay=[int, float], post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def signum(system_dict, learning_rate, momentum=0, weight_decay=0, nesterov=False):
    '''
    Select SIGNUM optimizer

    Args:
        system_dict (dict): System dictionary storing experiment state and set variables
        learning_rate (float): Initial base learning rate
        momentum (float): Momentum value for driving the weights towards minima
        weight_decay (float): Value for regularizing weights post every update

    Returns:
        dict: updated system dict
    '''
    system_dict["local"]["optimizer"] = "signum";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "signum";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["momentum"] = momentum;
    return system_dict;


@accepts(dict, [int, float], betas=tuple, epsilon=[int, float], weight_decay=[int, float], amsgrad=bool, post_trace=False)
#@TraceFunction(trace_args=False, trace_rv=False)
def ftml(system_dict, learning_rate, betas=(0.9, 0.999), epsilon=1e-08, weight_decay=0, amsgrad=False):
    '''
    Inactive function
    '''
    system_dict["local"]["optimizer"] = "ftml";
    system_dict["hyper-parameters"]["learning_rate"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["name"] = "ftml";
    system_dict["hyper-parameters"]["optimizer"]["params"]["lr"] = learning_rate;
    system_dict["hyper-parameters"]["optimizer"]["params"]["betas"] = betas;
    system_dict["hyper-parameters"]["optimizer"]["params"]["eps"] = epsilon;
    system_dict["hyper-parameters"]["optimizer"]["params"]["weight_decay"] = weight_decay;
    system_dict["hyper-parameters"]["optimizer"]["params"]["amsgrad"] = amsgrad;
    return system_dict;